
pc.stable.backend = function(x, cluster = NULL, whitelist, blacklist, test,
  alpha, B, debug = FALSE) {

  nodes = names(x)
  mb = structure(vector(length(nodes), mode = "list"), names = nodes)
  node.pairs = apply(subsets(nodes, 2), 1, function(x) list(arc = x))

  # find out which nodes are adjacent.
  for (dsep.size in seq(from = 0, to = length(nodes) - 2)) {

    node.pairs = smartSapply(cluster, node.pairs, pc.heuristic, data = x,
                   nodes = nodes, alpha = alpha, B = B, whitelist = whitelist,
                   blacklist = blacklist, test = test, dsep.size = dsep.size,
                   debug = debug)

  }#FOR

  # learn the skeleton.
  skeleton = do.call(rbind, lapply(node.pairs, function(x) { 

    if (x$p.value < alpha)
      return(x$arc)
    else
      return(NULL)

  }))

  # start encoding the bn object.
  skeleton = cache.structure(nodes, arcs.rbind(skeleton, skeleton, reverse2 = TRUE))
  # attach the d-separating sets.
  attr(skeleton, "dsep.set") = node.pairs

  return(skeleton)

}#PC.STABLE.OPTIMIZED

pc.heuristic = function(pair, data, nodes, alpha, B, whitelist, blacklist,
    test, dsep.size, debug = FALSE) {

  arc = pair$arc

  # check whether the arc is blacklisted in both directions (so that we do not
  # include it) or whitelisted in at least one direction (so that we include it).
  if (is.whitelisted(whitelist, arc, either = TRUE))
    return(list(arc = arc, p.value = 0))
  else if (is.blacklisted(blacklist, arc, both = TRUE))
    return(list(arc = arc, p.value = 1))

  # check whether the nodes are already d-separated.
  if (!is.null(pair$dsep.set))
    return(pair)

  if (debug) {

    cat("----------------------------------------------------------------\n")
    cat("* investigating", arc[1], "-", arc[2] ,
        ", d-separating sets of size", dsep.size, ".\n")

  }#THEN

  a = allsubs.test(x = arc[1], y = arc[2], sx = setdiff(nodes, arc),
        min = dsep.size, max = dsep.size, data = data, test = test,
        alpha = alpha, B = B, debug = debug)

  return(list(arc = arc, p.value = a["p.value"], dsep.set = attr(a, "dsep.set")))

}#PC.HEURISTIC

